
import UIKit

enum HipsterElementType: String {
  case Glasses1 = "glasses1"
  case Glasses2 = "glasses2"
  case Fedora1 = "fedora1"
}

struct HipsterElement: Printable {
  var type: HipsterElementType
  var center: CGPoint
  var width: CGFloat
  var transform: CGAffineTransform

  var description: String {
    return "\(type.rawValue) - center \(center) width \(width)"
  }
}

class Hipsterize {
  let image: UIImage

  init(image: UIImage) {
    self.image = Hipsterize.normalizedImage(image)
  }

  func processedImage() -> UIImage {
    let imageView = UIImageView(image: image)
    let ciImage = CIImage(image: image)

    let context = CIContext()
    let detector = CIDetector(ofType: CIDetectorTypeFace, context: context, options: [CIDetectorAccuracy: CIDetectorAccuracyHigh])
    let faces = detector.featuresInImage(ciImage)

    var counter = 0
    for face in faces as [CIFaceFeature] {
      let faceBounds = convertedRectForImage(image, originalRect: face.bounds)
      let overlayWidth = CGRectGetWidth(face.bounds)
      var eyeAngle: CGFloat = 0

      if (face.hasLeftEyePosition && face.hasRightEyePosition) {
        let rightEyePosition = convertedPointForImage(image, originalPoint: face.rightEyePosition)
        let leftEyePosition = convertedPointForImage(image, originalPoint: face.leftEyePosition)
        eyeAngle = atan2(rightEyePosition.y - leftEyePosition.y, rightEyePosition.x - leftEyePosition.x)

        let eyeCenter = CGPoint(x: (rightEyePosition.x + leftEyePosition.x) / 2.0, y: (rightEyePosition.y + leftEyePosition.y) / 2.0)

        counter += Int(arc4random_uniform(2))
        let glassesView = UIImageView(image: UIImage(named: counter % 2 == 0 ? "glasses1" : "glasses2"))
        glassesView.contentMode = .ScaleAspectFit
        glassesView.frame = CGRect(x: 0, y: 0, width: overlayWidth, height: overlayWidth / (glassesView.image!.size.width/glassesView.image!.size.height))
        glassesView.center = eyeCenter
        glassesView.transform = CGAffineTransformMakeRotation(eyeAngle)
        imageView.addSubview(glassesView)
      }

      counter += Int(arc4random_uniform(2))
      let fedoraView = UIImageView(image: UIImage(named: counter % 2 == 0 ? "fedora1" : "fedora2"))
      fedoraView.contentMode = .ScaleAspectFit
      fedoraView.frame = CGRect(x: 0, y: 0, width: overlayWidth*1.2, height: overlayWidth*1.2 / (fedoraView.image!.size.width/fedoraView.image!.size.height))


      let rectTopCenter = CGPoint(x: faceBounds.origin.x + CGRectGetWidth(faceBounds)/2, y: faceBounds.origin.y - CGRectGetHeight(fedoraView.frame)/3)
      let faceCenter = CGPoint(x: faceBounds.origin.x + CGRectGetWidth(faceBounds)/2, y: faceBounds.origin.y + CGRectGetHeight(faceBounds)/2)

      let s = sin(eyeAngle)
      let c = cos(eyeAngle)
      let rotatedPoint = CGPoint(x: (c * (rectTopCenter.x-faceCenter.x) - s * (rectTopCenter.y-faceCenter.y)) + faceCenter.x,
        y: (s * (rectTopCenter.x-faceCenter.x) + c * (rectTopCenter.y-faceCenter.y)) + faceCenter.y)

      fedoraView.center = rotatedPoint
      fedoraView.transform = CGAffineTransformMakeRotation(eyeAngle)
      imageView.addSubview(fedoraView)
    }

    UIGraphicsBeginImageContextWithOptions(image.size, false, image.scale)
    imageView.layer.renderInContext(UIGraphicsGetCurrentContext())
    let processedImage = UIGraphicsGetImageFromCurrentImageContext()
    UIGraphicsEndImageContext()
    return processedImage
  }

  private func convertedRectForImage(image: UIImage, originalRect: CGRect) -> CGRect {
    var transform = CGAffineTransformMakeScale(1, -1)
    transform = CGAffineTransformTranslate(transform, 0, -image.size.height);

    return CGRectApplyAffineTransform(originalRect, transform)
  }

  private func convertedPointForImage(image: UIImage, originalPoint: CGPoint) -> CGPoint {
    var convertedPoint = originalPoint
    var imageWidth = image.size.width
    var imageHeight = image.size.height

    switch (image.imageOrientation) {
    case .Up:
      convertedPoint.x = originalPoint.x;
      convertedPoint.y = imageHeight - originalPoint.y;
    case .Down:
      convertedPoint.x = imageWidth - originalPoint.x;
      convertedPoint.y = originalPoint.y;
    case .Left:
      convertedPoint.x = imageWidth - originalPoint.y;
      convertedPoint.y = imageHeight - originalPoint.x;
    case .Right:
      convertedPoint.x = originalPoint.y;
      convertedPoint.y = originalPoint.x;
    case .UpMirrored:
      convertedPoint.x = imageWidth - originalPoint.x;
      convertedPoint.y = imageHeight - originalPoint.y;
    case .DownMirrored:
      convertedPoint.x = originalPoint.x;
      convertedPoint.y = originalPoint.y;
    case .LeftMirrored:
      convertedPoint.x = imageWidth - originalPoint.y;
      convertedPoint.y = originalPoint.x;
    case .RightMirrored:
      convertedPoint.x = originalPoint.y;
      convertedPoint.y = imageHeight - originalPoint.x;
    }
    return convertedPoint
  }

  private class func normalizedImage(image: UIImage) -> UIImage {
    if image.imageOrientation == .Up {
      return image
    }

    UIGraphicsBeginImageContextWithOptions(image.size, false, image.scale)
    image.drawInRect(CGRect(origin: CGPoint(x: 0, y: 0), size: image.size))
    let normalizedImage = UIGraphicsGetImageFromCurrentImageContext()
    UIGraphicsEndImageContext()
    return normalizedImage
  }
}
